use std::{collections::HashMap, process::Stdio};

use htmxpress::Element;
use serde::{de::Unexpected, Deserialize, Deserializer, Serialize};
use tracing::warn;

use crate::error::NtitledError;

#[derive(Debug, Deserialize)]
pub struct FFProbeOutput {
    pub streams: Option<Vec<Stream>>,
}

impl FFProbeOutput {
    pub fn probe(path: &str) -> Result<Self, NtitledError> {
        let child = std::process::Command::new("ffprobe")
            .args([
                "-loglevel",
                "error",
                "-show_entries",
                "stream:stream_tags=language",
                "-show_streams",
                "-of",
                "json",
                path,
            ])
            .stdout(Stdio::piped())
            .stderr(Stdio::piped())
            .spawn()?
            .wait_with_output()?;

        let out = String::from_utf8(child.stdout)?;
        let err = String::from_utf8(child.stderr)?;
        let err = err.trim();

        if !err.is_empty() {
            warn!("{err}")
        }

        serde_json::from_str::<Self>(&out).map_err(NtitledError::from)
    }
}

#[derive(Debug, Serialize, Deserialize)]
#[serde(untagged)]
pub enum Stream {
    Video(VideoStream),
    Audio(AudioStream),
    Sub(SubtitleStream),
}

#[derive(Debug, Serialize, Deserialize, Element)]
pub struct VideoStream {
    pub index: usize,

    #[element("p")]
    #[format("Codec: {}")]
    #[serde(rename = "codec_name")]
    pub codec: String,

    #[element("p")]
    #[format("W: {}")]
    pub width: u16,

    #[element("p")]
    #[format("H: {}")]
    pub height: u16,

    #[element("p")]
    #[format("Captions: {}")]
    #[serde(deserialize_with = "bool_from_int")]
    pub closed_captions: bool,

    #[element("p")]
    #[format("Framerate: {}")]
    pub avg_frame_rate: String,

    #[serde(rename = "codec_long_name")]
    pub codec_long: String,
    pub profile: Option<String>,
    pub codec_type: String,
    pub coded_width: u16,
    pub coded_height: u16,

    pub start_time: String,
    pub disposition: Disposition,
    pub tags: Option<HashMap<String, String>>,
}

#[derive(Debug, Serialize, Deserialize)]
pub struct AudioStream {
    pub index: usize,

    #[serde(rename = "codec_name")]
    pub codec: String,
    #[serde(rename = "codec_long_name")]
    pub codec_long: String,

    pub sample_rate: String,
}

#[derive(Debug, Serialize, Deserialize, Element)]
pub struct SubtitleStream {
    pub index: usize,

    #[element("p")]
    #[format("Codec: {}")]
    #[serde(rename = "codec_name")]
    pub codec: String,

    #[element("p")]
    #[format("Duration: {}")]
    #[default("unknown")]
    pub duration: Option<String>,

    #[nest]
    pub tags: Tags,

    #[serde(rename = "codec_long_name")]
    pub codec_long: String,

    pub codec_type: String,
    pub avg_frame_rate: String,
    pub time_base: String,
    pub start_time: String,
    pub duration_ts: Option<usize>,
    pub disposition: Disposition,
}

#[derive(Debug, Serialize, Deserialize, Element)]
pub struct Tags {
    #[element("p")]
    #[format("Language: {}")]
    #[default("unspecified")]
    language: Option<String>,
}

#[derive(Debug, Serialize, Deserialize)]
pub struct Disposition {
    #[serde(deserialize_with = "bool_from_int")]
    pub default: bool,

    #[serde(deserialize_with = "bool_from_int")]
    pub dub: bool,

    #[serde(deserialize_with = "bool_from_int")]
    pub original: bool,

    #[serde(deserialize_with = "bool_from_int")]
    pub comment: bool,
}

fn bool_from_int<'de, D>(deserializer: D) -> Result<bool, D::Error>
where
    D: Deserializer<'de>,
{
    match u8::deserialize(deserializer)? {
        0 => Ok(false),
        1 => Ok(true),
        other => Err(serde::de::Error::invalid_value(
            Unexpected::Unsigned(other as u64),
            &"zero or one",
        )),
    }
}
