use axum::{
    response::Redirect,
    routing::{get, get_service},
    Router,
};
use ffmpeg::VideoStream;
use minijinja::Environment;
use ost::client::OSClient;
use serde::Serialize;
use std::{
    fs::{self, DirEntry},
    num::NonZeroUsize,
    path::Path,
    thread,
};
use tower_http::services::ServeDir;
use tracing::{debug, info, Level};

use crate::ffmpeg::{FFProbeOutput, SubtitleStream};

pub mod error;
pub mod ffmpeg;
pub mod htmx;
pub mod ost;
pub mod routes;

#[derive(Debug, Clone)]
pub struct State {
    client: OSClient,
    env: Environment<'static>,
    base_path: String,
}

lazy_static::lazy_static! {
    pub static ref INDEX: String =
        std::fs::read_to_string("assets/index.html").expect("missing template");
}

lazy_static::lazy_static! {
    pub static ref FILE_CONTAINER: String =
        std::fs::read_to_string("assets/filecontainer.html").expect("missing template");
}

lazy_static::lazy_static! {
    pub static ref MAX_THREADS: usize = std::thread::available_parallelism().unwrap_or(NonZeroUsize::new(1).unwrap()).into();
}

#[tokio::main]
async fn main() {
    tracing_subscriber::fmt()
        .with_max_level(Level::DEBUG)
        .init();

    info!("Starting ntitled");

    dotenv::dotenv().expect("could not load env");

    info!("Successfully loaded env");

    info!("Max threads: {}", *MAX_THREADS);

    let client = OSClient::init().await;

    info!("Successfully loaded OST client");

    let base_path = std::env::var("BASE_PATH").expect("base path not configured");
    let port = std::env::var("PORT").unwrap_or(String::from("3001"));

    let mut env = Environment::new();

    env.add_template("file_container", &FILE_CONTAINER)
        .expect("unable to add template");

    env.add_template("index", &INDEX)
        .expect("unable to add template");

    let state = State {
        client,
        env,
        base_path,
    };

    let router = axum::Router::new()
        .route("/", get(|| async { Redirect::permanent("/dir/root") }))
        .route("/dir/*key", get(routes::get_directory))
        .route("/subtitles", get(routes::search_subtitles))
        .route("/guess", get(routes::guess_it))
        .route("/download", get(routes::download_subtitles))
        .with_state(state);

    let router_static = Router::new().fallback(get_service(ServeDir::new("assets")));

    let listener = tokio::net::TcpListener::bind(format!("0.0.0.0:{port}"))
        .await
        .unwrap();

    info!("Starting server on 0.0.0.0:{port}");

    axum::serve(listener, router.merge(router_static))
        .await
        .unwrap();
}

pub fn scan_dir_contents(path: impl AsRef<Path>) -> std::io::Result<Vec<FileType>> {
    let entries = fs::read_dir(path)?
        .filter_map(Result::ok)
        .collect::<Vec<_>>();

    let subs = entries
        .iter()
        .filter_map(|entry| {
            let path = entry.path();
            let ext = path.extension()?.to_str()?;
            let (name, _) = path.file_name()?.to_str()?.split_once(ext)?;
            (ext == "srt").then_some(name.to_owned())
        })
        .collect::<Vec<_>>();

    let mut files = thread::scope(|scope| {
        let mut handles = vec![];

        let mut acc = Vec::with_capacity(*MAX_THREADS);

        for _ in 0..*MAX_THREADS {
            acc.push(vec![]);
        }

        let jobs = entries
            .into_iter()
            .enumerate()
            .fold(acc, |mut acc, (i, el)| {
                acc[i % *MAX_THREADS].push(el);
                acc
            });

        for entries in jobs {
            let task = scope.spawn(|| {
                let mut files = vec![];

                for entry in entries {
                    if let Some(file) = extract_entry(&subs, &entry) {
                        files.push(file)
                    }
                }

                files
            });

            handles.push(task);
        }

        let mut files = vec![];

        for handle in handles {
            match handle.join() {
                Ok(extracted_files) => {
                    files.extend(extracted_files);
                }
                Err(e) => {
                    debug!("Error while reading files: {e:?}")
                }
            }
        }

        files
    });

    files.sort();

    Ok(files)
}

fn extract_entry(subs: &[String], entry: &DirEntry) -> Option<FileType> {
    if entry.path().extension().is_some_and(|ext| {
        ext.to_str().is_some_and(|ext| ext == "srt") || ext.to_str().is_some_and(|ext| ext == "smi")
    }) {
        return None;
    }

    if entry.path().is_dir() {
        return Some(FileType::Directory(Directory {
            name: entry.file_name().to_str().unwrap_or_default().to_string(),
            path: entry.path().to_str().unwrap_or_default().to_string(),
        }));
        // Skip existing subtitles
    } else {
        let path = entry.path();

        let path_str = path.to_str()?;

        let out = match FFProbeOutput::probe(path_str) {
            Ok(streams) => streams,
            Err(e) => {
                debug!("File {path_str} is not a video: {e}");
                return None;
            }
        };

        let Some(streams) = out.streams else {
            return None;
        };

        let ext = path.extension()?.to_str()?;

        let (name, _) = path.file_name()?.to_str()?.split_once(ext)?;

        let subs_file = subs.contains(&name.to_string());

        let mut video_meta = vec![];
        let mut sub_meta = vec![];

        for stream in streams {
            match stream {
                ffmpeg::Stream::Video(s) => video_meta.push(s),
                ffmpeg::Stream::Sub(s) => sub_meta.push(s),
                ffmpeg::Stream::Audio(s) => {}
            }
        }

        let subs_embedded = !sub_meta.is_empty();

        Some(FileType::File(File {
            name: entry.file_name().to_str()?.to_string(),
            path: entry.path().to_str()?.to_string(),
            has_subs: subs_embedded || subs_file,
            video_meta,
            sub_meta,
        }))
    }
}

#[derive(Debug, Serialize, PartialEq, Eq, PartialOrd, Ord)]
pub enum FileType {
    Directory(Directory),
    File(File),
}

#[derive(Debug, Serialize)]
pub struct File {
    name: String,
    path: String,

    /// `true` if a same file exists with a `.srt` extension found
    /// in the same directory or if the file has embedded subs
    has_subs: bool,

    video_meta: Vec<VideoStream>,

    sub_meta: Vec<SubtitleStream>,
}

impl PartialEq for File {
    fn eq(&self, other: &Self) -> bool {
        self.name == other.name && self.path == other.path && self.has_subs == other.has_subs
    }
}

impl Eq for File {}

impl PartialOrd for File {
    fn partial_cmp(&self, other: &Self) -> Option<std::cmp::Ordering> {
        Some(self.cmp(other))
    }
}

impl Ord for File {
    fn cmp(&self, other: &Self) -> std::cmp::Ordering {
        match self.name.cmp(&other.name) {
            core::cmp::Ordering::Equal => {}
            ord => return ord,
        }
        match self.path.cmp(&other.path) {
            core::cmp::Ordering::Equal => {}
            ord => return ord,
        }
        self.has_subs.cmp(&other.has_subs)
    }
}

#[derive(Debug, Serialize, PartialEq, Eq, PartialOrd, Ord)]
pub struct Directory {
    name: String,

    path: String,
}
