use chrono::{DateTime, Utc};
use serde::{Deserialize, Serialize};

// SUBTITLES

#[derive(Debug, Deserialize, Serialize)]
pub struct SearchResponse {
    pub total_pages: usize,
    pub total_count: usize,
    pub per_page: usize,
    pub page: usize,
    pub data: Vec<Subtitle>,
}

/// https://opensubtitles.stoplight.io/docs/opensubtitles-api/573f76acc1493-subtitle
#[derive(Debug, Deserialize, Serialize)]
pub struct Subtitle {
    pub id: String,
    pub r#type: String,
    pub attributes: SubtitleAttrs,
}

/// https://opensubtitles.stoplight.io/docs/opensubtitles-api/573f76acc1493-subtitle
#[derive(Debug, Deserialize, Serialize)]
pub struct SubtitleAttrs {
    pub subtitle_id: String,
    pub language: Option<String>,
    pub download_count: usize,
    pub new_download_count: usize,
    pub hearing_impaired: bool,
    pub hd: bool,
    pub fps: f32,
    pub votes: usize,
    pub ratings: f32,
    pub from_trusted: bool,
    pub foreign_parts_only: bool,
    pub ai_translated: bool,
    pub release: String,
    pub comments: Option<String>,
    pub legacy_subtitle_id: usize,
    pub legacy_uploader_id: usize,
    pub uploader: Uploader,
    pub feature_details: Feature,
    pub url: String,
    pub moviehash_match: Option<bool>,
    pub points: Option<usize>,
    pub files: Vec<SubtitleFile>,
}

#[derive(Debug, Deserialize, Serialize)]
pub struct SubtitleFile {
    pub file_id: usize,
    pub cd_number: usize,
    pub file_name: String,
}

#[derive(Debug, Deserialize, Serialize)]
pub struct Uploader {
    uploader_id: Option<usize>,
    name: String,
    rank: String,
}

#[derive(Debug, Deserialize, Serialize)]
#[serde(untagged)]
pub enum Feature {
    TvShow(TvShow),
    Episode(Episode),
    Movie(Movie),
}

#[derive(Debug, Deserialize, Serialize)]
pub struct TvShow {
    pub title: String,
    pub original_title: String,
    pub year: usize,
    pub imdb_id: usize,
    pub tmdb_id: usize,
    pub feature_id: usize,
    pub url: String,
    pub img_url: String,
    pub subtitles_count: usize,
}

#[derive(Debug, Deserialize, Serialize)]
pub struct Episode {
    pub feature_id: usize,
    pub feature_type: String,
    pub year: usize,
    pub title: String,
    pub movie_name: String,
    pub imdb_id: usize,
    pub season_number: Option<usize>,
    pub episode_number: Option<usize>,
    pub parent_imdb_id: Option<usize>,
    pub parent_title: Option<String>,
    pub parent_feature_id: Option<usize>,
}

#[derive(Debug, Deserialize, Serialize)]
pub struct Movie {
    pub title: String,
    pub original_title: String,
    pub year: usize,
    pub subtitles_count: usize,
    pub seasons_count: usize,
    pub parent_title: String,
    pub season_number: Option<usize>,
    pub episode_number: Option<usize>,
    pub imdb_id: usize,
    pub tmdb_id: usize,
    pub feature_id: String,
    pub feature_type: String,
    pub url: String,
    pub img_url: String,
}

// AUTH

#[derive(Debug, Deserialize)]
pub struct LoginResponse {
    pub user: User,
    pub token: String,
    pub status: usize,
}

#[derive(Debug, Deserialize)]
pub struct User {
    pub allowed_downloads: usize,
    pub allowed_translations: usize,
    pub level: String,
    pub user_id: usize,
    pub ext_installed: bool,
    pub vip: bool,
}

#[derive(Debug, Serialize)]
pub struct Login<'a> {
    pub username: &'a str,
    pub password: &'a str,
}

// GUESS IT

#[derive(Debug, Deserialize)]
pub struct GuessIt {
    pub title: String,
    pub audio_codec: Option<String>,
    pub video_codec: Option<String>,
    pub release_group: Option<String>,
    pub r#type: Option<String>,

    // Available only for TV shows
    pub episode_title: Option<String>,
    pub season: Option<usize>,
    pub episode: Option<usize>,
}

// DOWNLOAD

#[derive(Debug, Deserialize)]
#[serde(untagged)]
pub enum DownloadResponse {
    Success(SubtitleFileInfo),
    Error { message: String },
}

#[derive(Debug, Deserialize)]
pub struct SubtitleFileInfo {
    pub link: String,
    pub file_name: String,
    pub requests: usize,
    pub remaining: usize,
    pub message: String,
    pub reset_time: String,
    pub reset_time_utc: DateTime<Utc>,
}

#[derive(Debug, Deserialize)]
pub struct SubtitleResponse {
    pub hashmatch: bool,
    pub title: String,
    pub original_title: Option<String>,
    pub year: usize,
    pub download_count: usize,
    pub parent_title: Option<String>,
    pub episode_number: Option<usize>,
    pub season_number: Option<usize>,
    pub files: Vec<SubtitleFile>,
    pub language: Option<String>,
}

impl From<Subtitle> for SubtitleResponse {
    fn from(value: Subtitle) -> Self {
        let SubtitleAttrs {
            feature_details,
            download_count,
            files,
            language,
            moviehash_match,
            ..
        } = value.attributes;

        match feature_details {
            crate::ost::data::Feature::TvShow(item) => {
                let TvShow {
                    title,
                    original_title,
                    year,
                    ..
                } = item;

                Self {
                    title,
                    original_title: Some(original_title),
                    year,
                    download_count,
                    parent_title: None,
                    episode_number: None,
                    season_number: None,
                    files,
                    language,
                    hashmatch: moviehash_match.is_some_and(|is| is),
                }
            }
            crate::ost::data::Feature::Episode(item) => {
                let Episode {
                    year,
                    title,
                    parent_title,
                    episode_number,
                    season_number,
                    ..
                } = item;
                Self {
                    title,
                    original_title: None,
                    year,
                    download_count,
                    parent_title,
                    episode_number,
                    season_number,
                    files,
                    language,
                    hashmatch: moviehash_match.is_some_and(|is| is),
                }
            }
            crate::ost::data::Feature::Movie(item) => {
                let Movie {
                    title,
                    original_title,
                    year,
                    parent_title,
                    episode_number,
                    season_number,
                    ..
                } = item;
                Self {
                    title,
                    original_title: Some(original_title),
                    year,
                    download_count,
                    parent_title: Some(parent_title),
                    episode_number,
                    season_number,
                    files,
                    language,
                    hashmatch: moviehash_match.is_some_and(|is| is),
                }
            }
        }
    }
}
